/* 
 * File:   final_main.c
 * Authors: Abdu, Vance, Sophia
 *
 * Created on: October 24, 2015, 4:05 PM
 * Last edit on: November 11, 2015
 */

#include "config.h"
// threading library
#include "pt_cornell_1_2_1.h"
// graphics libraries
#include "tft_master.h"
#include "tft_gfx.h"
#include "lsm303d.h"
#include <math.h>
#include <string.h>

/* Final Project Wooooo
 * 
 */

// === the fixed point macros ========================================
typedef signed int fix16 ;
#define multfix16(a,b) ((fix16)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16:16
#define float2fix16(a) ((fix16)((a)*65536.0)) // 2^16
#define fix2float16(a) ((float)(a)/65536.0)
#define fix2int16(a)    ((int)((a)>>16))
#define int2fix16(a)    ((fix16)((a)<<16))
#define divfix16(a,b) ((fix16)((((signed long long)(a)<<16)/(b)))) 
#define sqrtfix16(a) (float2fix16(sqrt(fix2float16(a)))) 
#define absfix16(a) abs(a)

//sine table
static int sinetable[256];

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_count, pt_serial, pt_lcd, pt_comp,
              pt_input, pt_dds, pt_buttons, pt_directions;
// system 1 second interval tick
int sys_time_seconds;

// ===  compass =====================
static short X;
static short Y;
int chk;
float h;

// ====== GPS + navigation variables ======================================
#define GP_max_chars 100
#define GGA_max_chars 66
static double longitude [8] = {42.44597, 42.44671, 42.44719, 42.44383, 42.44760, 42.44471, 42.44723, 42.44656};
static double latitude [8] = {76.48074, 76.48436, 76.48311, 76.48582, 76.48504, 76.48242, 76.48230, 76.48564};
static char name[8][25] = {"Barton Hall", "Cornell Store", "Day Hall", "Law School", "McGraw Tower", "Phillips-Duffield", "Uris Hall", "Willard Straight Hall"};
int onoff;
int degrees;
double minutes;
static double min2dec(double inmin){
    degrees=floor(inmin/100.0);
    minutes=100*(inmin/100.0-degrees);
    return degrees+minutes/60.0;
}
static float north, west, lastnorth[6], lastwest[6], test;

// ===== sound stuff ========//
// A-channel, 1x, active
#define DAC_config_chan_A 0b0011000000000000
#define DAC_config_chan_B 0xB000

//constants for Interaural Time Difference 
#define HeadRadius (float).089126768 //56/(200*3.14159)
#define SoundSpeed (float)300
#define ITDconstant (float)HeadRadius/SoundSpeed

// config.h sets 40 MHz
#define	SYS_FREQ 40000000

// ======= sound variables =========================================

volatile fix16 IIDattenuation;
volatile float IIDattFloat; 

//--- start DDS global variables--//
volatile unsigned int DAC_data_left, DAC_data_right ;// output value
volatile unsigned int accum1 = 0, accum2 = 0;
volatile unsigned int f1_incr, f2_incr;
volatile signed int PulseTimer1=0, PulseTimer2 = 0;
volatile signed int multiply_index1=0, multiply_index2=0;


volatile float ITDcycles, ITDtime; 
volatile unsigned int leftFast,rightFast; 
//volatile fix16 IIDattenuation;

volatile int ISRtimerValue;
static char buffer1[60];
//--- end DDS global variables--// 

//== Timer 3 interrupt handler ===========================================

void __ISR(_TIMER_3_VECTOR, ipl2) Timer3Handler(void)
{
    mT3ClearIntFlag();
    static int cycleNumber=0;
    static fix16 attenuation_index1, attenuation_index2; 

    
    if(rightFast){
        cycleNumber++;
        //implement time delay
        if (cycleNumber>ITDcycles){
            accum1+=f1_incr; //left

            //pulse the sinewave    
            if (PulseTimer1 <= 4000){
                multiply_index1 =0;
                PulseTimer1++;
            } else if (PulseTimer1<6000){
                multiply_index1 = 0; //1;
                attenuation_index1 = IIDattenuation; // make left side quieter
                PulseTimer1++;
            }else {
                //play = 0; 
                PulseTimer1 = 0;
                //multiply_index = 1;
            }
        }
            accum2+=f2_incr; //right

            //pulse the sinwave
            if (PulseTimer2 <= 4000){
                multiply_index2 =0;
                PulseTimer2++;
            }else if (PulseTimer2<6000){
                multiply_index2 = 1;
                attenuation_index2 = int2fix16(1);
                PulseTimer2++;
            }else {
                //play = 0; 
                PulseTimer2 = 0;
                //multiply_index = 1;
            }
    }
    else{
        cycleNumber++;
        //implement time delay
            accum1+=f1_incr;

            //pulse the sinewave    
            if (PulseTimer1 <= 4000){
                multiply_index1 =0;
                PulseTimer1++;
            } else if (PulseTimer1<6000){
                multiply_index1 = 1;
                attenuation_index1 = int2fix16(1);
                PulseTimer1++;
            }else {
                //play = 0; 
                PulseTimer1 = 0;
                //multiply_index = 1;
            }
        

        if (cycleNumber>ITDcycles){
            accum2+=f2_incr; // right
            // generate  ramp
            //pulse the sinwave
            if (PulseTimer2 <= 4000){
                multiply_index2 =0;
                PulseTimer2++;
            } else if (PulseTimer2<6000){
                multiply_index2 = 0; //1;
                attenuation_index2 = IIDattenuation; // make right side quieter
                PulseTimer2++;
            }else {
                //play = 0; 
                PulseTimer2 = 0;
                //multiply_index = 1;
            }
        }

    }
    
    DAC_data_left = multfix16(  (2047+ sinetable[ accum1 >> 24 ]+sinetable[ accum1 >> 24 ])*multiply_index1  ,  attenuation_index1  ); // for testing
    DAC_data_right = multfix16( (2047+ sinetable[ accum2 >> 24 ]+sinetable[ accum2 >> 24 ])*multiply_index2  ,  attenuation_index2  ); // for testing
    
    // CS low to start transaction
     mPORTBClearBits(BIT_4); // start transaction
    // test for ready
     while (TxBufFullSPI2());
     // write to spi2
     WriteSPI2(DAC_config_chan_A | DAC_data_left);
    // test for done
     while (SPI2STATbits.SPIBUSY); // wait for end of transaction
     // CS high
     mPORTBSetBits(BIT_4); // end transaction
     
//     delay_us(2);
//     
     // CS low to start transaction
     mPORTBClearBits(BIT_4); // start transaction
      // test for ready
     while (TxBufFullSPI2());
     // write to spi2
     WriteSPI2(DAC_config_chan_B | DAC_data_right);
    // test for done 
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
     // CS high
     mPORTBSetBits(BIT_4); // end transaction
     
     
     
     ISRtimerValue = ReadTimer3();
}


//====================================================================
// build a string from the UART2 /////////////
//////////////////////////////////////////////
char PT_term_buffer_GPS[GP_max_chars];
int n_char;
int PT_GetSerialBuffer_GPS(struct pt *pt)
{
    static char ch;
    // mark the beginning of the input thread
    PT_BEGIN(pt);

    n_char = 0;
    //UART2ClearAllErrors();
    while(n_char < GP_max_chars)
    {
        // get the character
        // yield until there is a valid character so that other
        // threads can execute
        UART2ClearAllErrors();
        PT_YIELD_UNTIL(pt, UARTReceivedDataIsAvailable(UART2));
       // while(!UARTReceivedDataIsAvailable(UART2)){};
        ch = UARTGetDataByte(UART2);

        // end line
        if(ch == '\r'){
            PT_term_buffer_GPS[n_char] = 0; // zero terminate the string
            break;
        }
        else if (ch == '$') {
            n_char = 0;
            PT_term_buffer_GPS[n_char++] = ch ;
        }
        else  {PT_term_buffer_GPS[n_char++] = ch ;}
    } 

    // kill this input thread, to allow spawning thread to execute
    PT_EXIT(pt);
    // and indicate the end of the thread
    PT_END(pt);
}

// === Thread Serial ======================================================
// The serial interface, allows for user-configurable parameters
//static char * token; static int parse_count;
//static const char d[] = ",";
static int fix_valid; static float N, E, gps_time;
static char GGA_buffer[GGA_max_chars];
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);

      while(1) {

            //spawn a thread to handle terminal input from GPS
            // the input thread waits for input
            // -- BUT does NOT block other threads
            // string is returned in "PT_term_buffer_GPS"
            //UART2ClearAllErrors();
            PT_SPAWN(pt, &pt_input, PT_GetSerialBuffer_GPS(&pt_input) );
            // returns when the thread dies
            // in this case, when an NMEA string is received
            if (PT_term_buffer_GPS[4] == 'G'){ // only extract GGA string
                strncpy(GGA_buffer, PT_term_buffer_GPS, GGA_max_chars);
                if (GGA_buffer[35] == '.') { // confirm validity of reading
                    fix_valid = 1;
                    sscanf(GGA_buffer, "$%*s,%f,%f,%*s,%f,%*s\r", 
                           &gps_time, &N, &E);
                } else {
                    fix_valid = 0;
                }
            
            }

            // never exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread serial

// === Thread LCD ======================================================
// Update LCD at 5 FPS, with control values
char buffer[60];
static PT_THREAD (protothread_lcd(struct pt *pt))
{
    PT_BEGIN(pt);
      while(1) {

            // yield time 1/5 second
            PT_YIELD_TIME_msec(1000) ;
            tft_fillRoundRect(0,0, 240, 56, 1, ILI9340_BLACK);// x,y,w,h,radius,color
            tft_setTextColor(ILI9340_YELLOW); tft_setTextSize(1);
            // draw sys_time
            tft_setCursor(0, 0);
            sprintf(buffer,"%d", sys_time_seconds);
            tft_writeString(buffer);
            
            tft_setCursor(0, 14);
            sprintf(buffer,"c: %d, x: %x, y: %x, h: %f", chk, X, Y, h);
            tft_writeString(buffer);
//            // draw M_S, D_S,  P, I, D
//            tft_setCursor(0, 14);
//            tft_writeString(GGA_buffer);
//            // draw S, P, I, D values        
//            tft_setCursor(0, 28);
//            tft_writeString("UTC, LAT, LONG, VAL");
//            tft_setCursor(0, 42);
//            sprintf(buffer,"%.3f, %.3f, %.3f, %d", gps_time, N, E, fix_valid);
//            tft_writeString(buffer);   

            // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread lcd

// === Thread Count ======================================================
// update a 1 second tick counter
static PT_THREAD (protothread_count(struct pt *pt))
{
    PT_BEGIN(pt);
      while(1) {
            PT_YIELD_TIME_msec(1000) ; // yield time 1 second
            sys_time_seconds++ ;
            // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread count

// === Thread Compass ======================================================
static PT_THREAD (protothread_comp(struct pt *pt))
{
    PT_BEGIN(pt);
      while(1) {
            PT_YIELD_TIME_msec(50) ; // yield time 1 second
            //while (SPI1STATbits.SPIBUSY);
            X = magnetometer(0);
            Y = magnetometer(1);
            h =  atan2(Y,X) * 180/M_PI;
            // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // thread count

static PT_THREAD (protothread_dds(struct pt *pt))
{
    PT_BEGIN(pt);    
        //local variables
        static int n=0, t_counter =0, sizeOfdigit = 0, num_to_play = 0;
        static float offsetAngle = 0; //1.57; // needs to be in radians
//        static int onoff = 1;
          
        while(1){
                  f1_incr = 91483000; f2_incr = 91483000;
                  //offsetAngle = direction; 
                  
                  //calculate time delay
                  ITDtime = .0002970892271*(offsetAngle + sin(offsetAngle));
                  ITDcycles = fabsf(ITDtime*40000);
                  leftFast = offsetAngle<0; //left is earlier then right
                  rightFast = offsetAngle>0; //right is earlier then left
                 
                  //calculate amplitude attenuation
                  IIDattenuation = float2fix16( fabsf(  cos( offsetAngle)  ) );
                  IIDattFloat = cos(offsetAngle);
             
                  //turn on/off sound
                  // enable/disable the core timer interrupt
                  // onoff == 1, then enable interrupt
                  INTEnable(INT_T3, onoff);
                  
          
          PT_YIELD_TIME_msec(1);
        } // END WHILE(1)
  PT_END(pt);
} // DDS thread

// ========= BUTTON DEBOUNCING ===========//
int PushState1, press1, PushState2, press2, switch1;
static PT_THREAD (protothread_buttons(struct pt *pt))
{
    EnablePullDownA(BIT_0|BIT_2|BIT_3);
    PT_BEGIN(pt);
    mPORTASetPinsDigitalIn(BIT_0|BIT_2|BIT_3);
      while(1) {
          PT_YIELD_TIME_msec(50) ;
          switch1 = mPORTAReadBits(BIT_0);
          press2 = mPORTAReadBits(BIT_2);
          press1 = mPORTAReadBits(BIT_3);
       switch (PushState1) {
       case NoPush: 
          if (press1) PushState1=MaybePush;
          else PushState1=NoPush;
          break;
       case MaybePush:
          if (press1) {
             PushState1=Pushed;
             selected++;
          }
          else PushState1=NoPush;
          break;
       case Pushed:  
          if (press1) PushState1=Pushed; 
          else PushState1=MaybeNoPush;    
          break;
       case MaybeNoPush:
          if (press1) PushState1=Pushed; 
          else PushState1=NoPush;    
          break;
            } // end case
       switch (PushState2) {
       case NoPush: 
          if (press2) PushState2=MaybePush;
          else PushState2=NoPush;
          break;
       case MaybePush:
          if (press2) {
             PushState2=Pushed;
             selected--;
          }
          else PushState2=NoPush;
          break;
       case Pushed:  
          if (press1) PushState2=Pushed; 
          else PushState2=MaybeNoPush;    
          break;
       case MaybeNoPush:
          if (press2) PushState2=Pushed; 
          else PushState2=NoPush;    
          break;
            }   
       if(switch1) onoff=1;
       else onoff=0;
          
      } // END WHILE(1)
  PT_END(pt);
} // thread button

// === Main  ======================================================
// set up UART, lcd, threads
// then schedule them as fast as possible

int main(void)
{
  // === config the uart, DMA, vref, timer5 ISR =============
  PT_setup();

  // === setup system wide interrupts  ====================
  INTEnableSystemMultiVectoredInt();
  
  // === the display ============================
  // init the display
  tft_begin();
  comp_begin();
  chk = comp_initialize();
  tft_fillScreen(ILI9340_BLACK);
  //240x320 vertical display
  tft_setRotation(0); // Use tft_setRotation(1) for 320x240
  
  // === sound setup ==========================
  // sine table initialization
  static int k;
  for (k=0; k< 256; k++){
      sinetable[k] = (int) (1023.0 * sin(6.283*(float)k/256));
  }
  
  // setup timer 3 and it's interrupts
  // sample rate @ 40 kSamples/sec
  OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_1, 1000);
  ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
  mT3ClearIntFlag(); // and clear the interrupt flag
  
  // SCK2 is pin 26
  // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
  PPSOutput(2, RPB5, SDO2);
  
  // set CS for DAC - RB4 (pin 14)
  mPORTBSetPinsDigitalOut(BIT_4);
  mPORTBSetBits(BIT_4);
  
  // divide Fpb by 2, configure the I/O ports. Not using SS in this example
  // 16 bit transfer CKP=1 CKE=1
  // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
  // For any given peripherial, you will need to match these
  SpiChnOpen(SPI_CHANNEL2, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | 
             SPI_OPEN_CKE_REV, 2);
  
  //while (1);
  // === now the threads ====================
  // init the threads
  PT_INIT(&pt_count);
  PT_INIT(&pt_comp);
  PT_INIT(&pt_lcd);
  PT_INIT(&pt_serial);

  // schedule the threads
  while(1) {
    // round robin
    PT_SCHEDULE(protothread_count(&pt_count));
    PT_SCHEDULE(protothread_comp(&pt_comp));
    PT_SCHEDULE(protothread_lcd(&pt_lcd));
    PT_SCHEDULE(protothread_serial(&pt_serial));
  }
} // main


